/*
    Copyright (C) 2002 Timo Sirainen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "common.h"
#include "temp-string.h"
#include "commands.h"

/* Returns status items, or -1 if error */
static MailboxStatusItems get_status_items(Client *client, ImapArgList *list)
{
	const char *item;
	MailboxStatusItems items;

	items = 0;
	for (; list != NULL; list = list->next) {
		if (list->arg.type != IMAP_ARG_ATOM) {
			/* list may contain only atoms */
			client_send_command_error(client, "Status list "
						  "contains non-atoms.");
			return -1;
		}

		str_ucase(list->arg.data.str);
		item = list->arg.data.str;

		if (strcmp(item, "MESSAGES") == 0)
			items |= STATUS_MESSAGES;
		else if (strcmp(item, "RECENT") == 0)
			items |= STATUS_RECENT;
		else if (strcmp(item, "UIDNEXT") == 0)
			items |= STATUS_UIDNEXT;
		else if (strcmp(item, "UIDVALIDITY") == 0)
			items |= STATUS_UIDVALIDITY;
		else if (strcmp(item, "UNSEEN") == 0)
			items |= STATUS_UNSEEN;
		else {
			client_send_tagline(client, t_strconcat(
				"BAD Invalid status item ", item, NULL));
			return -1;
		}
	}

	return items;
}

static int get_mailbox_status(Client *client, const char *mailbox,
			      MailboxStatusItems items, MailboxStatus *status)
{
	Mailbox *box;

	if (strcasecmp(mailbox, "inbox") == 0)
		mailbox = "INBOX";

	if (client->mailbox != NULL &&
	    strcmp(client->mailbox->name, mailbox) == 0) {
		/* this mailbox is selected */
		box = client->mailbox;
	} else {
		/* open the mailbox */
		box = client->storage->open_mailbox(client->storage,
						    mailbox, FALSE);
		if (box == NULL)
			return FALSE;
	}

	if (!box->get_status(box, items, status))
		return FALSE;

	if (box != client->mailbox)
		box->close(box);

	return TRUE;
}

int cmd_status(Client *client)
{
	ImapArg *args;
	MailboxStatus status;
	MailboxStatusItems items;
	const char *mailbox;
	TempString *str;

	/* <mailbox> <status items> */
	if (!client_read_args(client, 2, 0, &args))
		return FALSE;

	mailbox = imap_arg_string(&args[0]);
	if (mailbox == NULL || args[1].type != IMAP_ARG_LIST) {
		client_send_command_error(client, "Status items must be list.");
		return TRUE;
	}

	/* get the items client wants */
	items = get_status_items(client, args[1].data.list);
	if (items == -1) {
		/* error */
		return TRUE;
	}

	/* get status */
	if (!get_mailbox_status(client, mailbox, items, &status)) {
		client_send_storage_error(client);
		return TRUE;
	}

	str = t_string_new(128);
	t_string_printfa(str, "* STATUS %s (", mailbox);
	if (items & STATUS_MESSAGES)
		t_string_printfa(str, "MESSAGES %u ", status.messages);
	if (items & STATUS_RECENT)
		t_string_printfa(str, "RECENT %u ", status.recent);
	if (items & STATUS_UIDNEXT)
		t_string_printfa(str, "UIDNEXT %u ", status.uidnext);
	if (items & STATUS_UIDVALIDITY)
		t_string_printfa(str, "UIDVALIDITY %u ", status.uidvalidity);
	if (items & STATUS_UNSEEN)
		t_string_printfa(str, "UNSEEN %u ", status.unseen);

	if (str->str[str->len-1] == ' ')
		t_string_truncate(str, str->len-1);
	t_string_append_c(str, ')');

	client_send_line(client, str->str);
	client_send_tagline(client, "OK Status completed.");

	return TRUE;
}
