/*
    Copyright (C) 2002 Timo Sirainen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "lib.h"
#include "index-storage.h"
#include "mail-modifylog.h"

int index_storage_sync(Mailbox *box, unsigned int *messages, int expunge,
		       MailExpungeFunc expunge_func,
		       MailFlagUpdateFunc flag_func,
		       void *user_data)
{
	IndexMailbox *ibox = (IndexMailbox *) box;
	ModifyLogRecord *log;
	MailIndexRecord *rec;
	MailFlags flags;
	unsigned int count;
	int failed;

	if (expunge && box->readonly) {
		mail_storage_set_error(box->storage, "Mailbox is read-only");
		return FALSE;
	}

	*messages = 0;

	if (!ibox->index->set_lock(ibox->index, expunge ?
				   MAIL_LOCK_EXCLUSIVE : MAIL_LOCK_SHARED))
		return mail_storage_set_index_error(ibox);

	/* show the log */
	log = mail_modifylog_get_nonsynced(ibox->index->modifylog, &count);
	if (log == NULL)
		failed = TRUE;

	for (; count > 0; count--, log++) {
		switch (log->type) {
		case RECORD_TYPE_EXPUNGE:
			if (expunge_func != NULL) {
				expunge_func(box, log->seq,
					     log->uid, user_data);
			}
			break;
		case RECORD_TYPE_FLAGS_CHANGED:
			if (flag_func == NULL)
				break;

			rec = ibox->index->lookup_uid_range(ibox->index,
							    log->uid,
							    log->uid);
			if (rec != NULL) {
				flags = rec->msg_flags;
				if (rec->uid >= ibox->index->first_recent_uid)
					flags |= MAIL_RECENT;

				flag_func(box, log->seq, log->uid,
					  flags, user_data);
			}
			break;
		}
	}

	/* mark synced */
	failed = !mail_modifylog_mark_synced(ibox->index->modifylog);

	if (!failed && expunge) {
		/* expunge messages */
		failed = !index_storage_expunge_locked(ibox, expunge_func,
						       user_data);
	}

	/* get the messages count even if there was some failures.
	   also it must be done after expunging messages */
	count = ibox->index->get_header(ibox->index)->messages_count;
	if (count != ibox->synced_messages_count) {
		if (count > ibox->synced_messages_count) {
			/* new messages in mailbox */
			*messages = count;
		}
		ibox->synced_messages_count = count;
	}

	if (!ibox->index->set_lock(ibox->index, MAIL_LOCK_UNLOCK) || failed)
		return mail_storage_set_index_error(ibox);
	return TRUE;
}
