/*
    Copyright (C) 2002 Timo Sirainen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "lib.h"
#include "index-storage.h"
#include "mail-messageset.h"

typedef struct {
	Mailbox *box;
	MailFlags flags;
	ModifyType modify_type;
	MailFlagUpdateFunc func;
	void *user_data;
} UpdateData;

static int update_func(MailIndex *index, MailIndexRecord *rec,
		       unsigned int seq, void *user_data)
{
	UpdateData *data = user_data;
	MailFlags flags;

	switch (data->modify_type) {
	case MODIFY_ADD:
		flags = rec->msg_flags | data->flags;
		break;
	case MODIFY_REMOVE:
		flags = rec->msg_flags & ~data->flags;
		break;
	case MODIFY_REPLACE:
		flags = data->flags;
		break;
	default:
		flags = 0;
		i_assert(0);
	}

	if (rec->uid >= index->first_recent_uid)
		flags |= MAIL_RECENT;

	if (!index->update_flags(index, rec, seq, flags, FALSE))
		return FALSE;

	if (data->func != NULL)
		data->func(data->box, seq, rec->uid, flags, data->user_data);
	return TRUE;
}

int index_storage_update_flags(Mailbox *box, const char *messageset, int uidset,
			       MailFlags flags, ModifyType modify_type,
			       MailFlagUpdateFunc func, void *user_data,
			       int *all_found)
{
	IndexMailbox *ibox = (IndexMailbox *) box;
        UpdateData data;
	int ret;

	if (box->readonly) {
		mail_storage_set_error(box->storage, "Mailbox is read-only");
		return FALSE;
	}

	if (!ibox->index->set_lock(ibox->index, MAIL_LOCK_EXCLUSIVE))
		return mail_storage_set_index_error(ibox);

	data.box = box;
	data.flags = flags & ~MAIL_RECENT; /* \Recent can't be changed */
	data.modify_type = modify_type;
	data.func = func;
	data.user_data = user_data;

	if (uidset) {
		ret = mail_index_uidset_foreach(ibox->index, messageset,
						ibox->synced_messages_count,
						update_func, &data);
	} else {
		ret = mail_index_messageset_foreach(ibox->index,
						    messageset,
						    ibox->synced_messages_count,
						    update_func, &data);
	}

	if (!ibox->index->set_lock(ibox->index, MAIL_LOCK_UNLOCK) || ret == -1)
		return mail_storage_set_index_error(ibox);

	if (all_found != NULL)
		*all_found = ret == 1;
	return TRUE;
}
