/*
 temp-mempool.c : Memory pool for temporary memory allocations

    Copyright (c) 2001-2002 Timo Sirainen

    Permission is hereby granted, free of charge, to any person obtaining
    a copy of this software and associated documentation files (the
    "Software"), to deal in the Software without restriction, including
    without limitation the rights to use, copy, modify, merge, publish,
    distribute, sublicense, and/or sell copies of the Software, and to
    permit persons to whom the Software is furnished to do so, subject to
    the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
    OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
    IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
    CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include <stdlib.h>

#include "lib.h"
#include "temp-mempool.h"

/* #define TEMP_POOL_DISABLE */

#define DEFAULT_STACK_SIZE (1024*16)
#define MIN_BLOCK_SIZE 1024

typedef struct _MemBlock {
	struct _MemBlock *next;

	unsigned int size, left;
	unsigned char data[1];
} MemBlock;

/* mem_stack contains last 10 t_push()ed blocks. After that new
   MemBlockStack is created and it's ->prev is set to mem_stack. */
#define MEM_LIST_BLOCK_COUNT 10
typedef struct _MemBlockStack {
	MemBlock *block[MEM_LIST_BLOCK_COUNT];
        int block_space_used[MEM_LIST_BLOCK_COUNT];
        struct _MemBlockStack *prev;
} MemBlockStack;

static MemBlock *mem_last, *unused_blocks, *biggest_unused;
static MemBlockStack *mem_stack, *unused_stack;

static MemBlock *last_buffer_block;
static unsigned int last_buffer_size;

#ifdef TEMP_POOL_DISABLE
static void *last_buffer;
#endif

int t_push(void)
{
        MemBlockStack *stack;
	int n;

        stack = NULL;
	for (n = 0; n < MEM_LIST_BLOCK_COUNT; n++) {
		if (mem_stack->block[n] == NULL)
                        break;
	}

	if (n == MEM_LIST_BLOCK_COUNT) {
		n = 0;
		if (unused_stack == NULL) {
			stack = calloc(sizeof(MemBlockStack), 1);
			if (stack == NULL)
				i_panic("t_push(): Out of memory");
		} else {
			stack = unused_stack;
			unused_stack = unused_stack->prev;
                        memset(stack->block, 0, sizeof(stack->block));
		}
		stack->prev = mem_stack;
		mem_stack = stack;
	}

	mem_stack->block[n] = mem_last;
	mem_stack->block_space_used[n] = mem_last->left;

        return n;
}

static void update_biggest_unused(MemBlock *block)
{
	unsigned int biggest;

	biggest = biggest_unused == NULL ? 0 : biggest_unused->size;
	while (block != NULL) {
		if (block->size > biggest) {
			biggest = block->size;
			biggest_unused = block;
		}
                block = block->next;
	}
}

int t_pop(void)
{
	MemBlockStack *stack;
	MemBlock *block, *last_block;
        int n;

	if (mem_stack->block[0] == NULL) {
		if (mem_stack->prev == NULL) {
			i_warning("t_pop() called with empty stack");
                        return -1;
		}

		stack = mem_stack->prev;
		mem_stack->prev = unused_stack;
		unused_stack = mem_stack;
                mem_stack = stack;
	}

	for (n = 0; n < MEM_LIST_BLOCK_COUNT-1; n++) {
		if (mem_stack->block[n+1] == NULL)
                        break;
	}

	/* block is now partially used */
	block = mem_stack->block[n];
	mem_stack->block[n] = NULL;

	if (block->next != NULL) {
		/* block->next and everything after it is moved to
		   beginning of unused_blocks list */
		update_biggest_unused(block->next);

		last_block = block;
		while (last_block->next != NULL)
			last_block = last_block->next;

		last_block->next = unused_blocks;
		unused_blocks = block->next;
		block->next = NULL;
	}

	block->left = mem_stack->block_space_used[n];
	mem_last = block;

        return n;
}

static inline MemBlock *find_prev_unused(MemBlock *block)
{
	MemBlock *list;

	list = unused_blocks;
	while (list->next != block)
		list = list->next;
        return list;
}

static inline MemBlock *mem_block_alloc(int size)
{
	MemBlock *block;
	int alloc_size;

	if (size < MIN_BLOCK_SIZE)
                size = MIN_BLOCK_SIZE;

	alloc_size = nearest_power(size);
	block = malloc(sizeof(MemBlock)-1 + alloc_size);
	if (block == NULL) {
		i_panic("mem_block_alloc(): "
			"Out of memory when allocating %d bytes",
			sizeof(MemBlock)-1 + alloc_size);
	}
	block->size = alloc_size;
	block->next = NULL;
        return block;
}

static void *t_malloc_int(unsigned int size, int permanent)
{
	MemBlock *block;
        void *ret;

        last_buffer_block = NULL;
	if (size == 0)
		return NULL;

	if (mem_last->left >= size) {
		ret = mem_last->data + (mem_last->size - mem_last->left);
                if (permanent)
			mem_last->left -= size;
                else
			last_buffer_block = mem_last;
		return ret;
	}

        block = NULL;
	if (unused_blocks != NULL) {
		/* check if we find something from unused_blocks list.. */
		if (unused_blocks->size >= size) {
			block = unused_blocks;
			if (permanent)
				unused_blocks = unused_blocks->next;
		} else if (biggest_unused != NULL &&
			   biggest_unused->size >= size) {
			block = biggest_unused;
			if (permanent) {
				find_prev_unused(block)->next =
					biggest_unused->next;
			}
		}

		if (permanent && block == biggest_unused) {
			biggest_unused = NULL;
			update_biggest_unused(unused_blocks);
		}
	}

	if (block == NULL) {
		block = mem_block_alloc(size);
		if (!permanent) {
			update_biggest_unused(block);

			block->next = unused_blocks;
			unused_blocks = block;
		}
	}

	if (!permanent) {
                block->left = block->size;
		last_buffer_block = block;
		return block->data;
	}

	block->left = block->size - size;
	block->next = NULL;
	mem_last->next = block;
	mem_last = block;

        return mem_last->data;
}

void *t_malloc(unsigned int size)
{
#ifndef TEMP_POOL_DISABLE
        return t_malloc_int(size, TRUE);
#else
	(void)t_malloc_int(size, TRUE);
	return malloc(size);
#endif
}

void *t_malloc0(unsigned int size)
{
	void *mem;

#ifndef TEMP_POOL_DISABLE
	mem = t_malloc_int(size, TRUE);
#else
	(void)t_malloc_int(size, TRUE);
	mem = malloc(size);
#endif
	memset(mem, 0, size);
        return mem;
}

void *t_buffer_get(unsigned int size)
{
        last_buffer_size = size;
#ifndef TEMP_POOL_DISABLE
	return t_malloc_int(size, FALSE);
#else
	(void)t_malloc_int(size, FALSE);
	last_buffer = malloc(size);
	return last_buffer;
#endif
}

void *t_buffer_reget(void *buffer, unsigned int size)
{
	int old_size;
        void *new_buffer;

	old_size = last_buffer_size;
	if (size <= old_size)
                return buffer;

	new_buffer = t_buffer_get(size);
	if (new_buffer != buffer)
                memcpy(new_buffer, buffer, old_size);

        return new_buffer;
}

void t_buffer_alloc(unsigned int size)
{
        MemBlock **block;

	i_assert(last_buffer_block != NULL);
	i_assert(last_buffer_size >= size);

	if (size == 0)
		return;

	last_buffer_block->left -= size;
	if (last_buffer_block == mem_last) {
		/* last allocated block */
                return;
	}

        /* need to make this block allocated */
	for (block = &unused_blocks; *block != NULL; block = &(*block)->next) {
		if (*block == last_buffer_block) {
			*block = (*block)->next;
			break;
		}
	}

	if (last_buffer_block == biggest_unused) {
		biggest_unused = NULL;
		update_biggest_unused(unused_blocks);
	}

	last_buffer_block->next = NULL;
	mem_last->next = last_buffer_block;
	mem_last = last_buffer_block;
	last_buffer_block = NULL;
}

void temp_mempool_init(void)
{
	mem_stack = calloc(sizeof(MemBlockStack), 1);
        if (mem_stack == NULL)
		i_panic("temp_mempool_init(): Out of memory");

	mem_last = mem_block_alloc(DEFAULT_STACK_SIZE);
	mem_last->left = mem_last->size;
	mem_last->next = NULL;

        last_buffer_block = NULL;
	last_buffer_size = 0;
}

void temp_mempool_deinit(void)
{
	while (unused_blocks != NULL) {
                MemBlock *block = unused_blocks;
		unused_blocks = unused_blocks->next;

                free(block);
	}

	while (unused_stack != NULL) {
                MemBlockStack *stack = unused_stack;
		unused_stack = unused_stack->prev;

                free(stack);
	}

	free(mem_last);
	free(mem_stack);
}
