/*
    Copyright (C) 2002 Timo Sirainen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "common.h"
#include "lib-signals.h"
#include "network.h"

#include "auth-process.h"
#include "login-process.h"

#include <unistd.h>
#include <syslog.h>
#include <sys/stat.h>
#include <sys/wait.h>

#define IMAP2_PORT 143 /* imap2 / imap4 */
#define IMAPS_PORT 993 /* IMAP over SSL */

const char *process_names[PROCESS_TYPE_MAX] = {
	"unknown",
	"auth",
	"login",
	"imap"
};

static IOLoop ioloop;
static Timeout to_children;

HashTable *pids;
int imap_fd;

int validate_str(const char *str, int max_len)
{
	int i;

	for (i = 0; i < max_len; i++) {
		if (str[i] == '\0')
			return TRUE;
	}

	return FALSE;
}

void clean_child_process(void)
{
	extern char **environ;

	/* remove all environment, we don't need them */
	if (environ != NULL)
		environ[0] = NULL;

	(void)close(imap_fd);
	closelog();
}

static void sig_quit(int signo)
{
	io_loop_stop(ioloop);
}

static void children_check_timeout(void *user_data, Timeout timeout)
{
	const char *process_type_name;
	pid_t pid;
	int status, process_type;

	while ((pid = waitpid(-1, &status, WNOHANG)) > 0) {
		/* get the type and remove from hash */
		process_type = PID_GET_PROCESS_TYPE(pid);
		PID_REMOVE_PROCESS_TYPE(pid);

		/* write errors to syslog */
		process_type_name = process_names[process_type];
		if (WIFEXITED(status)) {
			status = WEXITSTATUS(status);
			if (status != 0) {
				i_error("child %d (%s) returned error %d",
					pid, process_type_name, status);
			}
		} else if (WIFSIGNALED(status)) {
			i_error("child %d (%s) killed with signal %d",
				pid, process_type_name, WTERMSIG(status));
		}
	}

	if (pid == -1 && errno != EINTR && errno != ECHILD)
		i_warning("waitpid() failed: %s", strerror(errno));
}

static void imap_socket_open(IPADDR *iface)
{
	int port;

	port = IMAP2_PORT;
	imap_fd = net_listen(iface, &port);
	if (imap_fd == -1)
		i_fatal("listen() failed: %s", strerror(errno));
}

static void main_init(void)
{
	/* open the syslog immediately so chroot() won't break logging */
	openlog("imap-master", LOG_NDELAY, LOG_MAIL);

	i_set_panic_handler(i_syslog_panic_handler);
	i_set_fatal_handler(i_syslog_fatal_handler);
        i_set_error_handler(i_syslog_error_handler);
        i_set_warning_handler(i_syslog_warning_handler);

	lib_init_signals(sig_quit);
	settings_read();

	/* deny file access from everyone else except owner */
        (void)umask(0077);

	pids = hash_create(default_pool, 128, NULL, NULL);

	imap_socket_open(NULL);
	to_children = timeout_add(100, children_check_timeout, NULL);

	auth_processes_init();
	login_processes_init();
}

static void main_deinit(void)
{
        if (lib_signal_kill != 0)
		i_warning("Killed with signal %d", lib_signal_kill);

	login_processes_deinit();
	auth_processes_deinit();

	timeout_remove(to_children);
	net_disconnect(imap_fd);

	hash_destroy(pids);
	closelog();
}

static void daemonize(void)
{
	pid_t pid;

	pid = fork();
	if (pid < 0)
		i_fatal("fork() failed: %s", strerror(errno));

	if (pid != 0)
		_exit(0);
}

int main(int argc, char *argv[])
{
	lib_init();

	if (argc != 2 || strcmp(argv[1], "-F") != 0)
		daemonize();

	ioloop = io_loop_create();

	main_init();
        io_loop_run(ioloop);
	main_deinit();

	io_loop_destroy(ioloop);
	lib_deinit();

        return 0;
}
